/*
 * Sixth 3D engine. Author: Svjatoslav Agejenko.
 * This project is released under Creative Commons Zero (CC0) license.
 */
package eu.svjatoslav.sixth.e3d.renderer.octree.raytracer;

import eu.svjatoslav.sixth.e3d.geometry.Point3D;
import eu.svjatoslav.sixth.e3d.gui.Avatar;
import eu.svjatoslav.sixth.e3d.math.Transform;
import eu.svjatoslav.sixth.e3d.renderer.raster.Color;
import eu.svjatoslav.sixth.e3d.renderer.raster.shapes.basic.line.LineAppearance;
import eu.svjatoslav.sixth.e3d.renderer.raster.shapes.composite.TexturedRectangle;
import eu.svjatoslav.sixth.e3d.renderer.raster.texture.Texture;

import javax.imageio.ImageIO;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.net.URL;

/**
 * Camera that sees the raytraced scene.
 * It is represented on the scene as a textured rectangle.
 */
public class Camera extends TexturedRectangle {

    public static final int SIZE = 100;
    public static final int IMAGE_SIZE = 500;
    private final CameraView cameraView;

    public Camera(final Avatar avatar, final double zoom) {
        super(new Transform(avatar.getLocation().clone()));
        cameraView = new CameraView(avatar, zoom);

        computeCameraCoordinates(avatar);

        addWaitNotification(getTexture());
    }

    private void addWaitNotification(final Texture texture) {
        // add hourglass icon
        try {
            final BufferedImage sprite = getSprite("eu/svjatoslav/sixth/e3d/examples/hourglass.png");
            texture.graphics.drawImage(sprite, IMAGE_SIZE / 2,
                    (IMAGE_SIZE / 2) - 30, null);
        } catch (final Exception ignored) {
        }

        // add "Please wait..." message
        texture.graphics.setColor(java.awt.Color.WHITE);
        texture.graphics.setFont(new Font("Monospaced", Font.PLAIN, 10));
        texture.graphics.drawString("Please wait...", (IMAGE_SIZE / 2) - 20,
                (IMAGE_SIZE / 2) + 30);
    }

    private void computeCameraCoordinates(final Avatar avatar) {
        initialize(SIZE, SIZE, IMAGE_SIZE, IMAGE_SIZE, 3);

        Point3D cameraCenter = new Point3D();

        topLeft.setValues(cameraCenter.x, cameraCenter.y, cameraCenter.z + SIZE);
        topRight.clone(topLeft);
        bottomLeft.clone(topLeft);
        bottomRight.clone(topLeft);

        final float viewAngle = (float) .6;

        topLeft.rotate(cameraCenter, -viewAngle, -viewAngle);
        topRight.rotate(cameraCenter, viewAngle, -viewAngle);
        bottomLeft.rotate(cameraCenter, -viewAngle, viewAngle);
        bottomRight.rotate(cameraCenter, viewAngle, viewAngle);

        topLeft.rotate(cameraCenter, -avatar.getAngleXZ(), -avatar.getAngleYZ());
        topRight.rotate(cameraCenter, -avatar.getAngleXZ(), -avatar.getAngleYZ());
        bottomLeft.rotate(cameraCenter, -avatar.getAngleXZ(), -avatar.getAngleYZ());
        bottomRight.rotate(cameraCenter, -avatar.getAngleXZ(), -avatar.getAngleYZ());

        final Color cameraColor = new Color(255, 255, 0, 255);
        final LineAppearance appearance = new LineAppearance(2, cameraColor);

        addShape(appearance.getLine(topLeft, topRight));
        addShape(appearance.getLine(bottomLeft, bottomRight));
        addShape(appearance.getLine(topLeft, bottomLeft));
        addShape(appearance.getLine(topRight, bottomRight));

    }

    public CameraView getCameraView() {
        return cameraView;
    }

    public BufferedImage getSprite(final String ref) throws IOException {
        final URL url = this.getClass().getClassLoader().getResource(ref);
        return ImageIO.read(url);
    }

}
